@@ -4,6 +4,14 @@ require 'redis-store'
 module ActiveSupport
   module Cache
     class RedisStore < Store
+
+      ERRORS_TO_RESCUE = [
+        Errno::ECONNREFUSED,
+        Errno::EHOSTUNREACH,
+        Redis::CannotConnectError,
+        Redis::ConnectionError
+      ].freeze
+
       attr_reader :data
 
       # Instantiate the store.
@@ -34,7 +42,8 @@ module ActiveSupport
       #     pool: ::ConnectionPool.new(size: 1, timeout: 1) { ::Redis::Store::Factory.create("localhost:6379/0") })
       #     # => supply an existing connection pool (e.g. for use with redis-sentinel or redis-failover)
       def initialize(*addresses)
-        @options = addresses.extract_options!
+        @options = addresses.dup.extract_options!
+        addresses = addresses.map(&:dup)
 
         @data = if @options[:pool]
                   raise "pool must be an instance of ConnectionPool" unless @options[:pool].is_a?(ConnectionPool)
@@ -45,9 +54,9 @@ module ActiveSupport
                   pool_options[:size]    = options[:pool_size] if options[:pool_size]
                   pool_options[:timeout] = options[:pool_timeout] if options[:pool_timeout]
                   @pooled = true
-                  ::ConnectionPool.new(pool_options) { ::Redis::Store::Factory.create(addresses) }
+                  ::ConnectionPool.new(pool_options) { ::Redis::Store::Factory.create(*addresses) }
                 else
-                  ::Redis::Store::Factory.create(addresses)
+                  ::Redis::Store::Factory.create(*addresses)
                 end
 
         super(@options)
@@ -57,7 +66,10 @@ module ActiveSupport
         options = merged_options(options)
         instrument(:write, name, options) do |payload|
           entry = options[:raw].present? ? value : Entry.new(value, options)
-          write_entry(namespaced_key(name, options), entry, options)
+          if options[:expires_in].present? && options[:race_condition_ttl].present? && options[:raw].blank?
+            options[:expires_in] = options[:expires_in].to_f + options[:race_condition_ttl].to_f
+          end
+          write_entry(normalize_key(name, options), entry, options)
         end
       end
 
@@ -78,7 +90,8 @@ module ActiveSupport
             with do |store|
               !(keys = store.keys(matcher)).empty? && store.del(*keys)
             end
-          rescue Errno::ECONNREFUSED, Redis::CannotConnectError
+          rescue *ERRORS_TO_RESCUE
+            raise if raise_errors?
             false
           end
         end
@@ -92,36 +105,52 @@ module ActiveSupport
       #   cache.read_multi "rabbit", "white-rabbit", :raw => true
       def read_multi(*names)
         options = names.extract_options!
-        keys = names.map{|name| namespaced_key(name, options)}
-        values = with { |c| c.mget(*keys) }
-        values.map! { |v| v.is_a?(ActiveSupport::Cache::Entry) ? v.value : v }
+        return {} if names == []
 
-        # Remove the options hash before mapping keys to values
-        names.extract_options!
+        keys = names.map{|name| normalize_key(name, options)}
+        args = [keys, options]
+        args.flatten!
 
-        result = Hash[keys.zip(values)]
-        result.reject!{ |k,v| v.nil? }
-        result
+        instrument(:read_multi, names) do |payload|
+          values = with { |c| c.mget(*args) }
+          values.map! { |v| v.is_a?(ActiveSupport::Cache::Entry) ? v.value : v }
+
+          Hash[names.zip(values)].reject{|k,v| v.nil?}.tap do |result|
+            payload[:hits] = result.keys if payload
+          end
+        end
+      rescue *ERRORS_TO_RESCUE
+        raise if raise_errors?
+        {}
       end
 
       def fetch_multi(*names)
-        results = read_multi(*names)
         options = names.extract_options!
-        fetched = {}
+        return {} if names == []
 
-        with do |c|
-          c.multi do
-            fetched = names.inject({}) do |memo, (name, _)|
-              key = namespaced_key(name, options)
-              memo[key] = results.fetch(key) do
-                value = yield name
+        results = read_multi(*names, options)
+        need_writes = {}
+
+        fetched = names.inject({}) do |memo, name|
+          memo[name] = results.fetch(name) do
+            value = yield name
+            need_writes[name] = value
+            value
+          end
+
+          memo
+        end
+
+        begin
+          with do |c|
+            c.multi do
+              need_writes.each do |name, value|
                 write(name, value, options)
-                value
               end
-
-              memo
             end
           end
+        rescue *ERRORS_TO_RESCUE
+          raise if raise_errors?
         end
 
         fetched
@@ -148,9 +177,10 @@ module ActiveSupport
       #
       #   cache.increment "rabbit"
       #   cache.read "rabbit", :raw => true       # => "1"
-      def increment(key, amount = 1)
+      def increment(key, amount = 1, options = {})
+        options = merged_options(options)
         instrument(:increment, key, :amount => amount) do
-          with{|c| c.incrby key, amount}
+          with{|c| c.incrby normalize_key(key, options), amount}
         end
       end
 
@@ -175,15 +205,16 @@ module ActiveSupport
       #
       #   cache.decrement "rabbit"
       #   cache.read "rabbit", :raw => true       # => "-1"
-      def decrement(key, amount = 1)
+      def decrement(key, amount = 1, options = {})
+        options = merged_options(options)
         instrument(:decrement, key, :amount => amount) do
-          with{|c| c.decrby key, amount}
-
+          with{|c| c.decrby normalize_key(key, options), amount}
         end
       end
 
       def expire(key, ttl)
-        with { |c| c.expire key, ttl }
+        options = merged_options(nil)
+        with { |c| c.expire normalize_key(key, options), ttl }
       end
 
       # Clear all the data from the store.
@@ -205,7 +236,7 @@ module ActiveSupport
       end
 
       def with(&block)
-        if @pooled
+        if defined?(@pooled) && @pooled
           @data.with(&block)
         else
           block.call(@data)
@@ -220,7 +251,8 @@ module ActiveSupport
         def write_entry(key, entry, options)
           method = options && options[:unless_exist] ? :setnx : :set
           with { |client| client.send method, key, entry, options }
-        rescue Errno::ECONNREFUSED, Redis::CannotConnectError
+        rescue *ERRORS_TO_RESCUE
+          raise if raise_errors?
           false
         end
 
@@ -229,7 +261,8 @@ module ActiveSupport
           if entry
             entry.is_a?(ActiveSupport::Cache::Entry) ? entry : ActiveSupport::Cache::Entry.new(entry)
           end
-        rescue Errno::ECONNREFUSED, Redis::CannotConnectError
+        rescue *ERRORS_TO_RESCUE
+          raise if raise_errors?
           nil
         end
 
@@ -239,11 +272,15 @@ module ActiveSupport
         # It's really needed and use
         #
         def delete_entry(key, options)
-          entry = with { |c| c.del key }
-        rescue Errno::ECONNREFUSED, Redis::CannotConnectError
+          with { |c| c.del key }
+        rescue *ERRORS_TO_RESCUE
+          raise if raise_errors?
           false
         end
 
+        def raise_errors?
+          !!@options[:raise_errors]
+        end
 
         # Add the namespace defined in the options to a pattern designed to match keys.
         #
@@ -252,14 +289,23 @@ module ActiveSupport
         # only for strings with wildcards.
         def key_matcher(pattern, options)
           prefix = options[:namespace].is_a?(Proc) ? options[:namespace].call : options[:namespace]
+
+          pattern = pattern.inspect[1..-2] if pattern.is_a? Regexp
+
           if prefix
-            raise "Regexps aren't supported, please use string with wildcards." if pattern.is_a?(Regexp)
             "#{prefix}:#{pattern}"
           else
             pattern
           end
         end
+
+      private
+
+        if ActiveSupport::VERSION::MAJOR < 5
+          def normalize_key(*args)
+            namespaced_key(*args)
+          end
+        end
     end
   end
 end
-
